@@ -1,14 +1,16 @@
 //! Implementation of the top level `*Event` derive macro.
 
 use proc_macro2::{Span, TokenStream};
-use quote::quote;
+use quote::{format_ident, quote};
 use syn::{
-    Data, DataStruct, DeriveInput, Field, Fields, FieldsNamed, Ident, Meta, MetaList, NestedMeta,
+    parse_quote, Data, DataStruct, DeriveInput, Field, Fields, FieldsNamed, GenericParam, Ident,
+    Meta, MetaList, NestedMeta,
 };
 
 use crate::{
     event_parse::{to_kind_variation, EventKind, EventKindVariation},
     import_ruma_events,
+    util::is_non_stripped_room_event,
 };
 
 /// Derive `Event` macro code generation.
@@ -40,23 +42,31 @@ pub fn expand_event(input: DeriveInput) -> syn::Result<TokenStream> {
         ));
     };
 
-    let serialize_impl = expand_serialize_event(&input, &var, &fields, &ruma_events);
-    let deserialize_impl = expand_deserialize_event(&input, &kind, &var, &fields, &ruma_events)?;
-    let conversion_impl = expand_from_into(&input, &kind, &var, &fields, &ruma_events);
+    let mut res = TokenStream::new();
 
-    let eq_impl = expand_eq_ord_event(&input, &fields);
+    res.extend(expand_serialize_event(&input, var, &fields, &ruma_events));
+    res.extend(expand_deserialize_event(&input, kind, var, &fields, &ruma_events)?);
 
-    Ok(quote! {
-        #conversion_impl
-        #serialize_impl
-        #deserialize_impl
-        #eq_impl
-    })
+    if var.is_sync() {
+        res.extend(expand_sync_from_into_full(&input, kind, var, &fields, &ruma_events));
+    }
+
+    if matches!(kind, EventKind::Message | EventKind::State)
+        && matches!(var, EventKindVariation::Full | EventKindVariation::Sync)
+    {
+        res.extend(expand_redact_event(&input, kind, var, &fields, &ruma_events));
+    }
+
+    if is_non_stripped_room_event(kind, var) {
+        res.extend(expand_eq_ord_event(&input));
+    }
+
+    Ok(res)
 }
 
 fn expand_serialize_event(
     input: &DeriveInput,
-    var: &EventKindVariation,
+    var: EventKindVariation,
     fields: &[Field],
     ruma_events: &TokenStream,
 ) -> TokenStream {
@@ -74,12 +84,6 @@ fn expand_serialize_event(
                         state.serialize_field("content", &self.content)?;
                     }
                 }
-            } else if name == "prev_content" {
-                quote! {
-                    if let Some(content) = self.prev_content.as_ref() {
-                        state.serialize_field("prev_content", content)?;
-                    }
-                }
             } else if name == "unsigned" {
                 quote! {
                     if !self.unsigned.is_empty() {
@@ -87,8 +91,20 @@ fn expand_serialize_event(
                     }
                 }
             } else {
-                quote! {
-                    state.serialize_field(stringify!(#name), &self.#name)?;
+                let name_s = name.to_string();
+                match &field.ty {
+                    syn::Type::Path(syn::TypePath { path: syn::Path { segments, .. }, .. })
+                        if segments.last().unwrap().ident == "Option" =>
+                    {
+                        quote! {
+                            if let Some(content) = self.#name.as_ref() {
+                                state.serialize_field(#name_s, content)?;
+                            }
+                        }
+                    }
+                    _ => quote! {
+                        state.serialize_field(#name_s, &self.#name)?;
+                    },
                 }
             }
         })
@@ -117,8 +133,8 @@ fn expand_serialize_event(
 
 fn expand_deserialize_event(
     input: &DeriveInput,
-    _kind: &EventKind,
-    var: &EventKindVariation,
+    _kind: EventKind,
+    var: EventKindVariation,
     fields: &[Field],
     ruma_events: &TokenStream,
 ) -> syn::Result<TokenStream> {
@@ -169,9 +185,9 @@ fn expand_deserialize_event(
                         EventKindVariation::Redacted | EventKindVariation::RedactedSync => {
                             ty = quote! { #ruma_events::RedactedUnsignedWithPrevContent };
                         }
-                        EventKindVariation::Stripped
-                        | EventKindVariation::Initial
-                        | EventKindVariation::RedactedStripped => unreachable!(),
+                        EventKindVariation::Stripped | EventKindVariation::Initial => {
+                            unreachable!()
+                        }
                     }
                 }
 
@@ -391,86 +407,152 @@ fn expand_deserialize_event(
     })
 }
 
-fn expand_from_into(
+fn expand_redact_event(
     input: &DeriveInput,
-    kind: &EventKind,
-    var: &EventKindVariation,
+    kind: EventKind,
+    var: EventKindVariation,
     fields: &[Field],
     ruma_events: &TokenStream,
-) -> Option<TokenStream> {
+) -> TokenStream {
     let ruma_identifiers = quote! { #ruma_events::exports::ruma_identifiers };
 
+    let redacted_type = kind.to_event_ident(var.to_redacted());
+    let redacted_content_trait =
+        format_ident!("{}Content", kind.to_event_ident(EventKindVariation::Redacted));
     let ident = &input.ident;
 
-    let (impl_generics, ty_gen, where_clause) = input.generics.split_for_impl();
-
-    let fields: Vec<_> = fields.iter().flat_map(|f| &f.ident).collect();
-
-    if let EventKindVariation::Sync | EventKindVariation::RedactedSync = var {
-        let full_struct = kind.to_event_ident(&var.to_full_variation());
-        Some(quote! {
-            #[automatically_derived]
-            impl #impl_generics From<#full_struct #ty_gen> for #ident #ty_gen #where_clause {
-                fn from(event: #full_struct #ty_gen) -> Self {
-                    let #full_struct { #( #fields, )* .. } = event;
-                    Self { #( #fields, )* }
-                }
-            }
-
-            #[automatically_derived]
-            impl #impl_generics #ident #ty_gen #where_clause {
-                /// Convert this sync event into a full event, one with a room_id field.
-                pub fn into_full_event(
-                    self,
-                    room_id: #ruma_identifiers::RoomId,
-                ) -> #full_struct #ty_gen {
-                    let Self { #( #fields, )* } = self;
-                    #full_struct {
-                        #( #fields, )*
-                        room_id,
-                    }
-                }
-            }
-        })
-    } else {
-        None
+    let mut generics = input.generics.clone();
+    if generics.params.is_empty() {
+        return TokenStream::new();
     }
-}
 
-fn expand_eq_ord_event(input: &DeriveInput, fields: &[Field]) -> Option<TokenStream> {
-    fields.iter().flat_map(|f| f.ident.as_ref()).any(|f| f == "event_id").then(|| {
-        let ident = &input.ident;
-        let (impl_gen, ty_gen, where_clause) = input.generics.split_for_impl();
+    assert_eq!(generics.params.len(), 1, "expected one generic parameter");
+    let ty_param = match &generics.params[0] {
+        GenericParam::Type(ty) => ty.ident.clone(),
+        _ => panic!("expected a type parameter"),
+    };
 
-        quote! {
-            #[automatically_derived]
-            impl #impl_gen ::std::cmp::PartialEq for #ident #ty_gen #where_clause {
-                /// This checks if two `EventId`s are equal.
-                fn eq(&self, other: &Self) -> ::std::primitive::bool {
-                    self.event_id == other.event_id
-                }
-            }
+    let where_clause = generics.make_where_clause();
+    where_clause.predicates.push(parse_quote! { #ty_param: #ruma_events::RedactContent });
+    where_clause.predicates.push(parse_quote! {
+        <#ty_param as #ruma_events::RedactContent>::Redacted:
+            #ruma_events::#redacted_content_trait
+    });
 
-            #[automatically_derived]
-            impl #impl_gen ::std::cmp::Eq for #ident #ty_gen #where_clause {}
+    let (impl_generics, ty_gen, where_clause) = generics.split_for_impl();
 
-            #[automatically_derived]
-            impl #impl_gen ::std::cmp::PartialOrd for #ident #ty_gen #where_clause {
-                /// This compares `EventId`s and orders them lexicographically.
-                fn partial_cmp(&self, other: &Self) -> ::std::option::Option<::std::cmp::Ordering> {
-                    self.event_id.partial_cmp(&other.event_id)
-                }
-            }
+    let fields = fields.iter().filter_map(|field| {
+        let ident = field.ident.as_ref().unwrap();
 
-            #[automatically_derived]
-            impl #impl_gen ::std::cmp::Ord for #ident #ty_gen #where_clause {
-                /// This compares `EventId`s and orders them lexicographically.
-                fn cmp(&self, other: &Self) -> ::std::cmp::Ordering {
-                    self.event_id.cmp(&other.event_id)
+        if ident == "content" || ident == "prev_content" {
+            None
+        } else if ident == "unsigned" {
+            Some(quote! {
+                unsigned:
+                    #ruma_events::RedactedUnsigned::new_because(::std::boxed::Box::new(redaction))
+            })
+        } else {
+            Some(quote! {
+                #ident: self.#ident
+            })
+        }
+    });
+
+    quote! {
+        #[automatically_derived]
+        impl #impl_generics #ruma_events::Redact for #ident #ty_gen #where_clause {
+            type Redacted =
+                #ruma_events::#redacted_type<<#ty_param as #ruma_events::RedactContent>::Redacted>;
+
+            fn redact(
+                self,
+                redaction: #ruma_events::room::redaction::SyncRoomRedactionEvent,
+                version: &#ruma_identifiers::RoomVersionId,
+            ) -> Self::Redacted {
+                let content = #ruma_events::RedactContent::redact(self.content, version);
+                #ruma_events::#redacted_type {
+                    content,
+                    #(#fields),*
                 }
             }
         }
-    })
+    }
+}
+
+fn expand_sync_from_into_full(
+    input: &DeriveInput,
+    kind: EventKind,
+    var: EventKindVariation,
+    fields: &[Field],
+    ruma_events: &TokenStream,
+) -> TokenStream {
+    let ruma_identifiers = quote! { #ruma_events::exports::ruma_identifiers };
+
+    let ident = &input.ident;
+    let full_struct = kind.to_event_ident(var.to_full());
+    let (impl_generics, ty_gen, where_clause) = input.generics.split_for_impl();
+    let fields: Vec<_> = fields.iter().flat_map(|f| &f.ident).collect();
+
+    quote! {
+        #[automatically_derived]
+        impl #impl_generics ::std::convert::From<#full_struct #ty_gen>
+            for #ident #ty_gen #where_clause
+        {
+            fn from(event: #full_struct #ty_gen) -> Self {
+                let #full_struct { #( #fields, )* .. } = event;
+                Self { #( #fields, )* }
+            }
+        }
+
+        #[automatically_derived]
+        impl #impl_generics #ident #ty_gen #where_clause {
+            /// Convert this sync event into a full event, one with a room_id field.
+            pub fn into_full_event(
+                self,
+                room_id: ::std::boxed::Box<#ruma_identifiers::RoomId>,
+            ) -> #full_struct #ty_gen {
+                let Self { #( #fields, )* } = self;
+                #full_struct {
+                    #( #fields, )*
+                    room_id,
+                }
+            }
+        }
+    }
+}
+
+fn expand_eq_ord_event(input: &DeriveInput) -> TokenStream {
+    let ident = &input.ident;
+    let (impl_gen, ty_gen, where_clause) = input.generics.split_for_impl();
+
+    quote! {
+        #[automatically_derived]
+        impl #impl_gen ::std::cmp::PartialEq for #ident #ty_gen #where_clause {
+            /// Checks if two `EventId`s are equal.
+            fn eq(&self, other: &Self) -> ::std::primitive::bool {
+                self.event_id == other.event_id
+            }
+        }
+
+        #[automatically_derived]
+        impl #impl_gen ::std::cmp::Eq for #ident #ty_gen #where_clause {}
+
+        #[automatically_derived]
+        impl #impl_gen ::std::cmp::PartialOrd for #ident #ty_gen #where_clause {
+            /// Compares `EventId`s and orders them lexicographically.
+            fn partial_cmp(&self, other: &Self) -> ::std::option::Option<::std::cmp::Ordering> {
+                self.event_id.partial_cmp(&other.event_id)
+            }
+        }
+
+        #[automatically_derived]
+        impl #impl_gen ::std::cmp::Ord for #ident #ty_gen #where_clause {
+            /// Compares `EventId`s and orders them lexicographically.
+            fn cmp(&self, other: &Self) -> ::std::cmp::Ordering {
+                self.event_id.cmp(&other.event_id)
+            }
+        }
+    }
 }
 
 /// CamelCase's a field ident like "foo_bar" to "FooBar".
